import os
from functools import partial
from typing import Callable

import openai


def openai_wrapper(text: str, *args, **kwargs):
    api_key = os.environ.get("OPENAI_API_KEY")
    openai_response = openai.Completion.create(
        api_key=api_key,
        prompt=text,
        *args,
        **kwargs,
    )
    return openai_response["choices"][0]["text"]


def get_llm_ask(llm_api: Callable, *args, **kwargs):
    if llm_api == openai.Completion.create:
        return partial(openai_wrapper, *args, **kwargs)
    else:
        raise ValueError("Only openai.Completion.create supported right now.")



import warnings
from typing import Any, Dict, List
from types import SimpleNamespace

from lxml import etree as ET


def get_validators(element: ET._Element, strict: bool = False) -> List["Validator"]:
    """Get the formatters for an element.

    Args:
        element: The XML element.
        strict: If True, raise an error if the element is not registered.

    Returns:
        A list of formatters.
    """

    from guardrails.validators import types_to_validators, validators_registry

    if 'format' not in element.attrib:
        return []

    provided_formatters = element.attrib['format'].split(';')
    registered_formatters = types_to_validators[element.tag]

    valid_formatters = []

    for formatter in provided_formatters:
        # Check if the formatter has any arguments.

        formatter = formatter.strip()

        args = []
        formatter_with_args = formatter.split(':')
        if len(formatter_with_args) > 1:
            assert len(formatter_with_args) == 2, (
                f"Formatter {formatter} has too many arguments.")
            formatter, args = formatter_with_args
            formatter = formatter.strip()
            args = [x.strip() for x in args.strip().split(' ')]

            for i, arg in enumerate(args):
                # If arg is enclosed within curly braces, then it is a python expression.
                if arg[0] == '{' and arg[-1] == '}':
                    args[i] = eval(arg[1:-1])

        if formatter not in registered_formatters:
            if strict:
                raise ValueError(
                    f"Formatter {formatter} is not valid for element {element.tag}.")
            else:
                warnings.warn(
                    f"Formatter {formatter} is not valid for element {element.tag}.")
            continue

        # See if the formatter has an associated on_fail method.
        on_fail = None
        on_fail_attr_name = f'on-fail-{formatter}'
        if on_fail_attr_name in element.attrib:
            on_fail = element.attrib[on_fail_attr_name]
            # TODO(shreya): Load the on_fail method.
            # This method should be loaded from an optional script given at the
            # beginning of a gxml file.

        formatter = validators_registry[formatter]
        valid_formatters.append(formatter(*args, on_fail=on_fail))

    return valid_formatters


class DataType:

    def __init__(self, validators: List, children: Dict[str, Any]) -> None:
        self.validators = validators
        self._children = children

    @classmethod
    def from_str(self, s: str) -> "DataType":
        """Create a DataType from a string."""
        raise NotImplementedError("Abstract method.")

    def validate(self, key: str, value: Any, schema: Dict) -> Dict:
        """Validate a value."""
        raise NotImplementedError("Abstract method.")

    def set_children(self, element: ET._Element):
        raise NotImplementedError("Abstract method.")

    @classmethod
    def from_xml(cls, element: ET._Element, strict: bool = False) -> "DataType":
        data_type = cls([], {})
        data_type.set_children(element)
        data_type.validators = get_validators(element, strict=strict)
        return data_type

    @property
    def children(self) -> SimpleNamespace:
        return SimpleNamespace(**self._children)

registry: Dict[str, DataType] = {}


# Create a decorator to register a type
def register_type(name: str):
    def decorator(cls: type):
        registry[name] = cls
        return cls
    return decorator


class ScalarType(DataType):
    def validate(self, key: str, value: Any, schema: Dict) -> Dict:
        """Validate a value."""
        for validator in self.validators:
            schema = validator.validate_with_correction(key, value, schema)

            if schema is None:
                # The outcome of validation was to refrain from answering.
                return None

            if key not in schema:
                # The key may have been filtered out by a previous validator.
                break
        return schema

    def set_children(self, element: ET._Element):
        for _ in element:
            raise ValueError("ScalarType data type must not have any children.")


class NonScalarType(DataType):
    pass


@register_type("string")
class String(ScalarType):
    """
    Element tag: `<string>`
    """

    @classmethod
    def from_str(self, s: str) -> "String":
        """Create a String from a string."""
        return s


@register_type("integer")
class Integer(ScalarType):
    """
    Element tag: `<integer>`
    """


@register_type("float")
class Float(ScalarType):
    """
    Element tag: `<float>`
    """


@register_type("date")
class Date(ScalarType):
    """
    Element tag: `<date>`
    """


@register_type("time")
class Time(ScalarType):
    """
    Element tag: `<time>`
    """


@register_type("email")
class Email(ScalarType):
    """
    Element tag: `<email>`
    """


@register_type("url")
class URL(ScalarType):
    """
    Element tag: `<url>`
    """


@register_type("pythoncode")
class PythonCode(ScalarType):
    """
    Element tag: `<pythoncode>`
    """



@register_type("percentage")
class Percentage(ScalarType):
    """
    Element tag: `<percentage>`
    """


@register_type("list")
class List(NonScalarType):
    """
    Element tag: `<list>`
    """

    def validate(self, key: str, value: Any, schema: Dict) -> Dict:
        # Validators in the main list data type are applied to the list overall.

        for validator in self.validators:
            schema = validator.validate_with_correction(key, value, schema)

            if schema is None:
                # The outcome of validation was to refrain from answering.
                return None

            if key not in schema:
                # The key may have been filtered out by a previous validator.
                # In this case, we don't need to validate the items in the list,
                # since the list itself is not present.
                return schema

        if len(self._children) == 0:
            return schema

        item_type = list(self._children.values())[0]

        # TODO(shreya): Edge case: List of lists -- does this still work?
        for i, item in enumerate(value):
            value = item_type.validate(i, item, value)

        return schema

    def set_children(self, element: ET._Element):
        idx = 0
        for child in element:
            idx += 1
            if idx > 1:
                # Only one child is allowed in a list data type.
                # The child must be the datatype that all items in the list
                # must conform to.
                raise ValueError("List data type must have exactly one child.")
            child_data_type = registry[child.tag]
            self._children["item"] = child_data_type.from_xml(child)


@register_type("object")
class Object(NonScalarType):
    """
    Element tag: `<object>`
    """

    def validate(self, key: str, value: Any, schema: Dict) -> Dict:
        # Validators in the main object data type are applied to the object overall.

        for validator in self.validators:
            schema = validator.validate_with_correction(key, value, schema)

            if schema is None:
                # The outcome of validation was to refrain from answering.
                return None

            if key not in schema:
                # The key may have been filtered out by a previous validator.
                # In this case, we don't need to validate the items in the object,
                # since the object itself is not present.
                return schema

        if len(self._children) == 0:
            return schema

        # Types of supported children
        # 1. key_type
        # 2. value_type
        # 3. List of keys that must be present

        # TODO(shreya): Implement key type and value type later

        # Check for required keys
        for child_key, child_data_type in self._children.items():
            # Value should be a dictionary
            # child_key is an expected key that the schema defined
            # child_data_type is the data type of the expected key
            value = child_data_type.validate(
                child_key,
                value.get(child_key, None),
                value
            )

        schema[key] = value

        return schema

    def set_children(self, element: ET._Element):
        for child in element:
            child_data_type = registry[child.tag]
            self._children[child.attrib["name"]] = child_data_type.from_xml(child)


# @register_type("key")
# class Key(DataType):
    # """
    # Element tag: `<string>`
    # """


# @register_type("value")
# class Value(DataType):
    # """
    # Element tag: `<string>`
    # """


# @register_type("item")
# class Item(DataType):
    # """
    # Element tag: `<string>`
    # """



import json
import logging
from copy import deepcopy
from typing import Any, Dict, List, Tuple, Callable

from eliot import start_action, to_file

import guardrails.utils.reask_utils as reask_utils
from guardrails.llm_providers import get_llm_ask
from guardrails.prompt import Prompt
from guardrails.output_schema import OutputSchema
from guardrails.utils.rail_utils import read_rail
from guardrails.utils.logs_utils import GuardHistory, GuardLogs, GuardState

logger = logging.getLogger(__name__)
to_file(open("guardrails.log", "w"))


class Guard:
    def __init__(
        self,
        schema: OutputSchema,
        base_prompt: Prompt,
        num_reasks: int = 1,
    ):
        self.output_schema = schema
        self.num_reasks = num_reasks
        self.guard_state = GuardState([])

        parsed_rail_copy = deepcopy(self.output_schema.parsed_rail)
        output_schema_prompt = reask_utils.extract_prompt_from_xml(parsed_rail_copy)
        self.base_prompt = base_prompt.format(output_schema=output_schema_prompt)

    @classmethod
    def from_rail(cls, rail_file: str) -> "Guard":
        """Create an Schema from an XML file."""
        output_schema, base_prompt, _ = read_rail(rail_file=rail_file)
        return cls(output_schema, base_prompt)

    @classmethod
    def from_rail_string(cls, rail_string: str) -> "Guard":
        """Create an Schema from an XML string."""
        output_schema, base_prompt, _ = read_rail(rail_string=rail_string)
        return cls(output_schema, base_prompt)

    def __call__(
        self, llm_api: Callable, prompt_params: Dict = None, *args, **kwargs
    ) -> Tuple[str, Dict]:
        """Outermost function that calls the LLM and validates the output.

        Args:
            llm_api: The LLM API to call (e.g. openai.Completion.create)
            prompt_params: The parameters to pass to the prompt.format() method.
            *args: Additional arguments to pass to the LLM API.
            **kwargs: Additional keyword arguments to pass to the LLM API.

        Returns:
            The raw text output from the LLM and the validated output.
        """
        with start_action(action_type="guard_call", prompt_params=prompt_params):
            prompt = self.base_prompt
            if prompt_params is None:
                prompt_params = {}
            prompt = self.base_prompt.format(**prompt_params)
            llm_ask = get_llm_ask(llm_api, *args, **kwargs)

            return self.ask_with_validation(prompt, llm_ask)

    def ask_with_validation(
        self, prompt: str, llm_ask: Callable
    ) -> Tuple[str, Dict]:
        """Ask a question, and validate the output."""

        with start_action(action_type="ask_with_validation", prompt=prompt):
            guard_history = self.validation_inner_loop(
                prompt, llm_ask, 0, self.output_schema
            )

            return (
                guard_history.output,
                guard_history.validated_response,
            )

    def validation_inner_loop(
        self,
        formatted_prompt: str,
        llm_ask: Callable,
        reask_ctr: int,
        output_schema: OutputSchema,
        guard_history: GuardHistory = None,
    ) -> GuardHistory:
        """
        Ask a question, and validate the output.

        Args:
            formatted_prompt: The prompt to send to the LLM.
            reask_ctr: The number of times the user has been reasked.

        Returns:
            The raw output from the LLM, the output as a dict, and the
            validated output.
        """

        if guard_history is None:
            guard_history = GuardHistory([])

        with start_action(
            action_type="validation_inner_loop", reask_ctr=reask_ctr
        ) as action:

            output = llm_ask(formatted_prompt)
            action.log(message_type="info", prompt=formatted_prompt, output=output)

            try:
                output_as_dict = json.loads(output)
                action.log(message_type="info", output_as_dict=output_as_dict)
                validated_response, reasks = self.validate_output(
                    output_as_dict, output_schema
                )
            except json.decoder.JSONDecodeError:
                validated_response = None
                output_as_dict = None
                reasks = []
                action.log(message_type="info", output_as_dict=output_as_dict)

            action.log(
                message_type="info",
                validated_response=reask_utils.reask_json_as_dict(validated_response),
                reasks=[r.__dict__ for r in reasks],
            )

            gd_log = GuardLogs(
                prompt=formatted_prompt,
                output=output,
                output_as_dict=output_as_dict,
                validated_response=validated_response,
                reasks=reasks,
            )

            guard_history = guard_history.push(gd_log)

            if len(reasks) and reask_ctr < self.num_reasks:

                reask_json = reask_utils.prune_json_for_reasking(validated_response)
                reask_prompt, reask_schema = reask_utils.get_reask_prompt(
                    self.output_schema.parsed_rail, reasks, reask_json
                )

                return self.validation_inner_loop(
                    reask_prompt, llm_ask, reask_ctr + 1, reask_schema, guard_history
                )

            self.guard_state = self.guard_state.push(guard_history)

            return guard_history

    def validate_output(
        self, output: Dict[str, Any], schema: OutputSchema
    ) -> Tuple[Dict[str, Any], List[reask_utils.ReAsk]]:
        """Validate a output against the schema.

        Args:
            output: The output to validate.

        Returns:
            Tuple, where the first element is the validated output, and the
            second element is a list of tuples, where each tuple contains the
            path to the reasked element, and the ReAsk object.
        """

        validated_response = deepcopy(output)

        for field, value in validated_response.items():
            if field not in schema:
                logger.debug(f"Field {field} not in schema.")
                continue

            validated_response = schema[field].validate(
                field, value, validated_response
            )

        reasks = reask_utils.gather_reasks(validated_response)

        return (validated_response, reasks)

    def __repr__(self):
        def _print_dict(d: Dict[str, Any], indent: int = 0) -> str:
            """Print a dictionary in a nice way."""

            s = ""
            for k, v in d.items():
                if isinstance(v, dict):
                    s += f"{k}:\n{_print_dict(v, indent=indent + 1)}"
                else:
                    s += f"{' ' * (indent * 4)}{k}: {v}\n"

            return s

        schema = _print_dict(self.output_schema)

        return f"Schema({schema})"



"""XML utilities."""
import re
from typing import Dict, List, Optional, Tuple

from lxml import etree as ET

from guardrails.datatypes import registry as types_registry
from guardrails.output_schema import OutputSchema
from guardrails.prompt import Prompt
from guardrails.utils.constants import constants
from guardrails.validators import Validator


def read_rail(
    rail_file: Optional[str] = None, rail_string: Optional[str] = None
) -> Tuple[OutputSchema, Prompt, Dict]:
    """Read an RAIL file.

    Args:
        rail_file: The path to the RAIL file.

    Returns:
        The root element of the RAIL file.
    """
    if rail_file is not None:
        with open(rail_file, "r") as f:
            xml = f.read()
    elif rail_string is not None:
        xml = rail_string
    else:
        raise ValueError("Must pass either rail_file or rail_string.")
    parser = ET.XMLParser(encoding="utf-8")
    parsed_rail = ET.fromstring(xml, parser=parser)

    if "version" not in parsed_rail.attrib or parsed_rail.attrib["version"] != "0.1":
        raise ValueError(
            "RAIL file must have a version attribute set to 0.1."
            "Change the opening <rail> element to: <rail version='0.1'>"
        )

    raw_output_schema = parsed_rail.find("output")
    if raw_output_schema is None:
        raise ValueError("RAIL file must contain a output-schema element.")
    output_schema = load_output_schema(raw_output_schema)

    prompt = parsed_rail.find("prompt")
    if prompt is None:
        raise ValueError("RAIL file must contain a prompt element.")
    prompt = load_prompt(prompt)

    script = parsed_rail.find("script")
    if script is not None:
        script = load_script(script)

    return output_schema, prompt, script


def load_output_schema(root: ET._Element) -> OutputSchema:
    """Validate parsed XML, create a prompt and a Schema object."""

    output = OutputSchema(parsed_rail=root)
    strict = False
    if "strict" in root.attrib and root.attrib["strict"] == "true":
        strict = True

    for child in root:
        if isinstance(child, ET._Comment):
            continue
        child_name = child.attrib["name"]
        child_data = types_registry[child.tag].from_xml(child, strict=strict)
        output[child_name] = child_data

    return output


def load_prompt(root: ET._Element) -> Prompt:
    text = root.text

    # Substitute constants by reading the constants file.
    # Regex to extract all occurrences of @<constant_name>
    matches = re.findall(r"@(\w+)", text)

    # Substitute all occurrences of @<constant_name> with the value of the constant.
    for match in matches:
        text = text.replace(f"@{match}", constants[match])

    prompt = Prompt(text)

    return prompt


def load_script(root: ET._Element) -> Dict[str, List[Validator]]:
    pass



import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Union, Any, Optional, Callable, Dict

from guardrails.datatypes import registry as types_registry
from guardrails.utils.reask_utils import ReAsk

validators_registry = {}
types_to_validators = defaultdict(list)


logger = logging.getLogger(__name__)


def register_validator(name: str, data_type: Union[str, List[str]]):
    """Register a validator for a data type."""

    def decorator(cls: type):
        """Register a validator for a data type."""

        nonlocal data_type
        if isinstance(data_type, str):
            if data_type == "all":
                data_type = list(types_registry.keys())
            else:
                data_type = [data_type]

        # Make sure that the data type string exists in the data types registry.
        for dt in data_type:
            if dt not in types_registry:
                raise ValueError(f"Data type {dt} is not registered.")

            types_to_validators[dt].append(name)

        validators_registry[name] = cls
        return cls

    return decorator


@dataclass
class EventDetail(BaseException):
    """Event detail."""

    key: str
    value: Any
    schema: Dict[str, Any]
    error_message: str
    fix_value: Any


class Validator:
    """Base class for validators."""

    def __init__(self, on_fail: Optional[Callable] = None):
        if on_fail is not None:
            if isinstance(on_fail, str):
                if on_fail == "filter":
                    on_fail = self.filter
                elif on_fail == "refrain":
                    on_fail = self.refrain
                elif on_fail == "noop":
                    on_fail = self.noop
                elif on_fail == "fix":
                    on_fail = self.fix
                elif on_fail == "reask":
                    on_fail = self.reask
                else:
                    raise ValueError(f"Unknown on_fail value: {on_fail}.")
            self.on_fail = on_fail
        else:
            self.on_fail = self.fix

    def validate_with_correction(self, key, value, schema) -> Dict:
        try:
            return self.validate(key, value, schema)
        except EventDetail as e:
            logger.debug(
                f"Validator {self.__class__.__name__} failed for {key} with error {e}."
            )
            return self.on_fail(e)

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate a value."""

        raise NotImplementedError

    def fix(self, error: EventDetail) -> Dict:
        """Debug the incorrect value."""

        error.schema[error.key] = error.fix_value
        return error.schema

    def reask(self, error: EventDetail) -> Dict:
        """Reask disambiguates the validation failure into a helpful error message."""

        error.schema[error.key] = ReAsk(error.value, error.error_message)
        return error.schema

    def filter(self, error: EventDetail) -> Dict:
        """If validation fails, filter the offending key from the schema."""

        logger.debug(f"Filtering {error.key} from schema...")

        schema = error.schema
        key = error.key
        value = error.value

        if isinstance(schema, dict):
            schema.pop(key)
        elif isinstance(schema, list):
            schema.remove(value)

        return schema

    def refrain(self, error: EventDetail) -> Optional[Dict]:
        """If validation fails, refrain from answering."""

        logger.debug(f"Refusing to answer {error.key}...")

        return None

    def noop(self, error: EventDetail) -> Dict:
        """If validation fails, do nothing."""

        logger.debug(
            f"Validator {self.__class__.__name__} failed for {error.key}, "
            "but doing nothing..."
        )

        return error.schema


# @register_validator('required', 'all')
# class Required(Validator):
#     """Validate that a value is not None."""

#     def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> bool:
#         """Validate that a value is not None."""

#         return value is not None


# @register_validator('description', 'all')
# class Description(Validator):
#     """Validate that a value is not None."""

#     def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> bool:
#         """Validate that a value is not None."""

#         return value is not None


@register_validator(name="valid-range", data_type=["integer", "float", "percentage"])
class ValidRange(Validator):
    """Validate that a value is within a range."""

    def __init__(
        self, min: int = None, max: int = None, on_fail: Optional[Callable] = None
    ):
        """Initialize the validator."""
        super().__init__(on_fail=on_fail)

        self._min = min
        self._max = max

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is within a range."""

        logger.debug(f"Validating {value} is in range {self._min} - {self._max}...")

        if self._min is not None and value < self._min:
            raise EventDetail(
                key,
                value,
                schema,
                f"Value {value} is less than {self._min}.",
                self._min,
            )

        if self._max is not None and value > self._max:
            raise EventDetail(
                key,
                value,
                schema,
                f"Value {value} is greater than {self._max}.",
                self._max,
            )

        return schema


@register_validator(name="valid-choices", data_type="all")
class ValidChoices(Validator):
    """Validate that a value is within a range."""

    def __init__(self, choices: List[Any], on_fail: Optional[Callable] = None):
        """Initialize the validator."""
        super().__init__(on_fail=on_fail)
        self._choices = choices

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is within a range."""

        logger.debug(f"Validating {value} is in choices {self._choices}...")

        if value not in self._choices:
            raise EventDetail(
                key,
                value,
                schema,
                f"Value {value} is not in choices {self._choices}.",
                None,
            )

        return schema


@register_validator(name="lower-case", data_type="string")
class LowerCase(Validator):
    """Validate that a value is lower case."""

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is lower case."""

        logger.debug(f"Validating {value} is lower case...")

        if not value.lower() == value:
            raise EventDetail(
                key,
                value,
                schema,
                f"Value {value} is not lower case.",
                value.lower(),
            )

        return schema


@register_validator(name="upper-case", data_type="string")
class UpperCase(Validator):
    """Validate that a value is upper case."""

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is upper case."""

        logger.debug(f"Validating {value} is upper case...")

        if not value.upper() == value:
            raise EventDetail(
                key,
                value,
                schema,
                f"Value {value} is not upper case.",
                value.upper(),
            )

        return schema


@register_validator(name="length", data_type=["string", "list", "object"])
class ValidLength(Validator):
    """Validate that the length of value is within the expected range."""

    def __init__(
        self, min: int = None, max: int = None, on_fail: Optional[Callable] = None
    ):
        """
        Args:
            min: The minimum length of the value.
            max: The maximum length of the value.
        """
        super().__init__(on_fail=on_fail)
        self._min = int(min) if min is not None else None
        self._max = int(max) if max is not None else None

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is within a range."""

        logger.debug(
            f"Validating {value} is in length range {self._min} - {self._max}..."
        )

        if self._min is not None and len(value) < self._min:
            logger.debug(f"Value {value} is less than {self._min}.")

            # Repeat the last character to make the value the correct length.
            corrected_value = value + value[-1] * (self._min - len(value))
            raise EventDetail(
                key,
                value,
                schema,
                f"Value has length less than {self._min}. "
                f"Please return a longer output, that is shorter than {self._max} characters.",
                corrected_value,
            )

        if self._max is not None and len(value) > self._max:
            logger.debug(f"Value {value} is greater than {self._max}.")
            raise EventDetail(
                key,
                value,
                schema,
                f"Value has length greater than {self._max}. "
                f"Please return a shorter output, that is shorter than {self._max} characters.",
                value[0 : self._max],
            )

        return schema


@register_validator(name="two-words", data_type="string")
class TwoWords(Validator):
    """Validate that a value is upper case."""

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is upper case."""
        logger.debug(f"Validating {value} is two words...")

        if not len(value.split()) == 2:
            raise EventDetail(
                key,
                value,
                schema,
                "must be exactly two words",
                " ".join(value.split()[0:2]),
            )

        return schema


@register_validator(name="one-line", data_type="string")
class OneLine(Validator):
    """Validate that a value is a single line or sentence."""

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is a single line or sentence."""
        logger.debug(f"Validating {value} is a single line...")

        if not len(value.splitlines()) == 1:
            raise EventDetail(
                key,
                value,
                schema,
                f"Value {value} is not a single line.",
                value.splitlines()[0],
            )

        return schema


@register_validator(name="valid-url", data_type=["string", "url"])
class ValidUrl(Validator):
    """Validate that a value is a valid URL."""

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is a valid URL."""
        logger.debug(f"Validating {value} is a valid URL...")

        import requests

        # Check that the URL exists and can be reached
        try:
            response = requests.get(value)
            if not response.status_code == 200:
                raise EventDetail(
                    key,
                    value,
                    schema,
                    f"URL {value} returned status code {response.status_code}",
                    None,
                )
        except requests.exceptions.ConnectionError:
            raise EventDetail(
                key,
                value,
                schema,
                f"URL {value} could not be reached",
                None,
            )

        return schema


@register_validator(name="bug-free", data_type="pythoncode")
class BugFree(Validator):
    """Validate that a value is not a bug."""

    def validate(self, key: str, value: Any, schema: Union[Dict, List]) -> Dict:
        """Validate that a value is not a bug."""
        logger.debug(f"Validating {value} is not a bug...")

        # The value represents a Python code snippet. We need to execute it and check if there are any bugs
        try:
            exec(value)
        except Exception as e:
            raise EventDetail(
                key,
                value,
                schema,
                e,
                None,
            )

        return schema


from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Union

from lxml import etree as ET
from rich.pretty import pretty_repr

from guardrails.utils.constants import constants


@dataclass
class ReAsk:
    incorrect_value: Any
    error_message: str
    path: List[Any] = None

    def __repr__(self) -> str:
        return pretty_repr(self)


def gather_reasks(output: Dict) -> List[ReAsk]:
    """
    Traverse output and gather all ReAsk objects.
    Response is a nested dictionary, where values can also be lists or
    dictionaries.
    Make sure to also grab the corresponding paths (including list index),
    and return a list of tuples.
    """
    reasks = []

    def _gather_reasks(output: Union[list, dict], path: List[str] = []):
        if isinstance(output, dict):
            iterable = output.items()
        elif isinstance(output, list):
            iterable = enumerate(output)
        else:
            raise ValueError(f"Expected dict or list, got {type(output)}")
        for field, value in iterable:
            if isinstance(value, ReAsk):
                value.path = path + [field]
                reasks.append(value)

            if isinstance(value, dict):
                _gather_reasks(value, path + [field])

            if isinstance(value, list):
                for idx, item in enumerate(value):
                    if isinstance(item, ReAsk):
                        item.path = path + [field, idx]
                        reasks.append(item)
                    else:
                        _gather_reasks(item, path + [field, idx])

    _gather_reasks(output)
    return reasks


def get_reasks_by_element(
    reasks: List[ReAsk],
    parsed_rail: ET._Element,
) -> Dict[ET._Element, List[tuple]]:
    """Cluster reasks by the XML element they are associated with."""

    reasks_by_element = defaultdict(list)

    for reask in reasks:
        path = reask.path
        # TODO: does this work for all cases?
        query = "."
        for part in path:
            if isinstance(part, int):
                query += "/*"
            else:
                query += f"/*[@name='{part}']"

        # Find the element
        element = parsed_rail.find(query)

        reasks_by_element[element].append(reask)

    return reasks_by_element


def get_pruned_tree(
    root: ET._Element,
    reask_elements: List[ET._Element] = None,
) -> ET._Element:
    """Prune tree of any elements that are not in `reasks`.

    Return the tree with only the elements that are keys of `reasks` and
    their parents. If `reasks` is None, return the entire tree. If an
    element is removed, remove all ancestors that have no children.

    Args:
        root: The XML tree.
        reasks: The elements that are to be reasked.

    Returns:
        The prompt.
    """
    if reask_elements is None:
        return root

    # Get all elements in `root`
    elements = root.findall(".//*")
    for element in elements:
        if (element not in reask_elements) and len(element) == 0:
            parent = element.getparent()
            parent.remove(element)

            # Remove all ancestors that have no children
            while len(parent) == 0:
                grandparent = parent.getparent()
                grandparent.remove(parent)
                parent = grandparent

    return root


def prune_json_for_reasking(json_object: Any) -> Dict:
    """Validated JSON is a nested dictionary where some keys may be ReAsk objects.

    This function prunes the validated JSON of any object that is not a ReAsk object.
    It also keeps all of the ancestors of the ReAsk objects.

    Args:
        json_object: The validated JSON.

    Returns:
        The pruned validated JSON.
    """
    if isinstance(json_object, list):
        pruned_list = []
        for item in json_object:
            pruned_output = prune_json_for_reasking(item)
            if pruned_output is not None:
                pruned_list.append(pruned_output)
        if len(pruned_list):
            return pruned_list
        return None
    elif isinstance(json_object, dict):
        pruned_json = {}
        for key, value in json_object.items():
            if isinstance(value, ReAsk):
                pruned_json[key] = value
            elif isinstance(value, dict):
                pruned_output = prune_json_for_reasking(value)
                if pruned_output is not None:
                    pruned_json[key] = pruned_output
            elif isinstance(value, list):
                pruned_list = []
                for item in value:
                    pruned_output = prune_json_for_reasking(item)
                    if pruned_output is not None:
                        pruned_list.append(pruned_output)
                if len(pruned_list):
                    pruned_json[key] = pruned_list

        if len(pruned_json):
            return pruned_json

        return None
    else:
        if isinstance(json_object, ReAsk):
            return json_object
        return None


def extract_prompt_from_xml(tree: ET._Element) -> str:
    """Extract the prompt from an XML tree.

    Args:
        tree: The XML tree.

    Returns:
        The prompt.
    """
    # From the element tree, remove any action attributes like 'on-fail-*'.
    # Filter any elements that are comments.
    for element in tree.iter():

        if isinstance(element, ET._Comment):
            continue

        for attr in list(element.attrib):
            if attr.startswith("on-fail-"):
                del element.attrib[attr]

    # Return the XML as a string.
    return ET.tostring(tree, encoding="unicode", method="xml")


def get_reask_prompt(
    parsed_rail, reasks: List[ReAsk], reask_json: Dict
) -> Tuple[str, ET._Element]:
    """Construct a prompt for reasking.

    Args:
        parsed_rail: The parsed RAIL.
        reasks: List of tuples, where each tuple contains the path to the
            reasked element, and the ReAsk object (which contains the error
            message describing why the reask is necessary).
        reask_json: Pruned JSON that contains only ReAsk objects.

    Returns:
        The prompt.
    """
    parsed_rail_copy = deepcopy(parsed_rail)

    # Get the elements that are to be reasked
    reask_elements = get_reasks_by_element(reasks, parsed_rail_copy)

    # Get the pruned JSON so that it only contains ReAsk objects
    # Get the pruned tree
    pruned_tree = get_pruned_tree(parsed_rail_copy, list(reask_elements.keys()))
    pruned_tree_string = extract_prompt_from_xml(pruned_tree)

    reask_prompt_template = (
        constants["high_level_reask_prompt"] + constants["complete_json_suffix"]
    )

    reask_prompt = reask_prompt_template.format(
        previous_response=pretty_repr(reask_json),
        output_schema=pruned_tree_string,
    )

    return reask_prompt, pruned_tree


def reask_json_as_dict(json: Dict) -> Dict:
    """If a ReAsk object exists in the JSON, return it as a dictionary."""

    def _reask_json_as_dict(json_object: Any) -> Any:
        if isinstance(json_object, dict):
            return {
                key: _reask_json_as_dict(value) for key, value in json_object.items()
            }
        elif isinstance(json_object, list):
            return [_reask_json_as_dict(item) for item in json_object]
        elif isinstance(json_object, ReAsk):
            return json_object.__dict__
        else:
            return json_object

    return _reask_json_as_dict(json)



import os

from lxml import etree as ET


class ConstantsContainer:
    def __init__(self):
        self._constants = {}
        self.fill_constants()

    def fill_constants(self) -> None:

        constants_file = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "constants.xml"
        )

        with open(constants_file, "r") as f:
            xml = f.read()

        parser = ET.XMLParser(encoding="utf-8")
        parsed_constants = ET.fromstring(xml, parser=parser)

        for child in parsed_constants:
            if isinstance(child, ET._Comment):
                continue
            if isinstance(child, str):
                continue

            constant_name = child.tag
            constant_value = child.text
            self._constants[constant_name] = constant_value

    def __getitem__(self, key):
        return self._constants[key]

    def __setitem__(self, key, value):
        self._constants[key] = value

    def __delitem__(self, key):
        del self._constants[key]

    def __iter__(self):
        return iter(self._constants)
    
    def __len__(self):
        return len(self._constants)
    
    def __contains__(self, key):
        return key in self._constants
    
    def __repr__(self):
        return repr(self._constants)
    
    def __str__(self):
        return str(self._constants)
    
    def items(self):
        return self._constants.items()
    
    def keys(self):
        return self._constants.keys()
    
    def values(self):
        return self._constants.values()


constants = ConstantsContainer()



# Set up __init__.py so that users can do from guardrails import Response, Schema, etc.

from guardrails.guardrails import Guard
from guardrails.utils import constants, docs_utils, rail_utils
from guardrails.validators import Validator, register_validator

__all__ = [
    "Guard",
    "Validator",
    "register_validator",
    "rail_utils",
    "constants",
    "docs_utils",
]


from typing import Dict, Optional
from types import SimpleNamespace


from lxml import etree as ET

from guardrails.datatypes import DataType


class OutputSchema:
    """Output schema class that holds a _schema attribute.
    """
    def __init__(
        self,
        parsed_rail: Optional[ET._Element] = None,
        schema: Optional[Dict[str, DataType]] = None,
    ) -> None:

        if schema is None:
            schema = {}

        self._schema = SimpleNamespace(**schema)
        self.parsed_rail = parsed_rail

    def __getitem__(self, key: str) -> DataType:
        return getattr(self._schema, key)
    
    def __setitem__(self, key: str, value: DataType) -> None:
        setattr(self._schema, key, value)

    def __getattr__(self, key: str) -> DataType:
        return getattr(self._schema, key)

    def __contains__(self, key: str) -> bool:
        return hasattr(self._schema, key)

    def items(self) -> Dict[str, DataType]:
        return vars(self._schema).items()



import re


class Prompt:

    def __init__(self, source: str):
        self.source = source

        # Get all the variable names in the source string.
        # Variable names are surounded by curly braces, and may optionally contain a colon and a type.
        self.variable_names = re.findall(r'\{([\w\d_]+)\}', self.source)

    def add_output_schema_to_prompt(self, output_schema: str):
        self.output_schema = output_schema

    def append_to_prompt(self, text: str):
        self.source += text

    def __str__(self) -> str:
        return self.source

    def get_prompt_variables(self):
        return self.variable_names

    def format(self, **kwargs):
        """Format the prompt using the given keyword arguments."""
        return self.source.format(**kwargs)

    def make_vars_optional(self):
        """Make all variables in the prompt optional."""
        for var in self.variable_names:
            self.source = self.source.replace(f"{{{var}}}", f"{{{var}:}}")


class PromptRepo:
    def __init__(self):
        self.prompts = []

    def add_prompt(self, prompt_name: str, prompt_template: Prompt):
        # Make sure that the prompt name is not already in the repo.
        for prompt in self.prompts:
            if prompt["name"] == prompt_name:
                raise ValueError(f"Prompt with name {prompt_name} already exists.")
        self.prompts.append({"name": prompt_name, "template": prompt_template})

    def get_prompts(self):
        """Return a list of all prompts in the repo."""
        return self.prompts

    def get_prompt(self, prompt_name: str):
        """Return the prompt template for the given prompt name. If the prompt name is not found, return None."""
        for prompt in self.prompts:
            if prompt["name"] == prompt_name:
                return prompt["template"]
        return None

    def update_prompt(self, prompt_name: str, prompt_template: Prompt):
        """Update the prompt template for the given prompt name. If the prompt name is not found, raise an error."""
        found_prompt = False
        for prompt in self.prompts:
            if prompt["name"] == prompt_name:
                prompt["template"] = prompt_template
                return
        if not found_prompt:
            raise ValueError(f"Prompt with name {prompt_name} not found.")
